@@ -5,7 +5,7 @@ BEGIN {
   $KiokuDB::Backend::DBI::AUTHORITY = 'cpan:NUFFIN';
 }
 BEGIN {
-  $KiokuDB::Backend::DBI::VERSION = '1.12';
+  $KiokuDB::Backend::DBI::VERSION = '1.13';
 }
 use Moose;
 
@@ -221,6 +221,28 @@ has storage => (
 
 sub _build_storage { shift->schema->storage }
 
+has for_update => (
+    isa => "Bool",
+    is  => "ro",
+    default => 1,
+);
+
+has _for_update => (
+    isa => "Bool",
+    is  => "ro",
+    lazy_build => 1,
+);
+
+sub _build__for_update {
+    my $self = shift;
+
+    return (
+        $self->for_update
+            and
+        $self->storage->sqlt_type =~ /^(?:MySQL|Oracle|PostgreSQL)$/
+    );
+}
+
 has columns => (
     isa => ArrayRef[ValidColumnName|HashRef],
     is  => "ro",
@@ -330,6 +352,8 @@ sub insert {
 
     return unless @entries;
 
+    my $g = $self->schema->txn_scope_guard;
+
     $self->insert_rows( $self->entries_to_rows(@entries) );
 
     # hopefully we're in a transaction, otherwise this totally sucks
@@ -349,6 +373,8 @@ sub insert {
 
         $self->update_index(\%gin_index);
     }
+
+    $g->commit;
 }
 
 sub entries_to_rows {
@@ -413,6 +439,8 @@ sub entry_to_row {
 sub insert_rows {
     my ( $self, $insert, $update, $dbic ) = @_;
 
+    my $g = $self->schema->txn_scope_guard;
+
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
@@ -462,6 +490,14 @@ sub insert_rows {
 
         $_->insert_or_update for @$dbic;
     });
+
+    $g->commit;
+}
+
+sub prepare_select {
+    my ( $self, $dbh, $stmt ) = @_;
+
+    $dbh->prepare_cached($stmt . ( $self->_for_update ? " FOR UPDATE" : "" ), {}, 3); # 3 = don't use if still Active
 }
 
 sub prepare_insert {
@@ -469,7 +505,7 @@ sub prepare_insert {
 
     my @cols = @{ $self->_ordered_columns };
 
-    my $ins = $dbh->prepare("INSERT INTO entries (" . join(", ", @cols) . ") VALUES (" . join(", ", ('?') x @cols) . ")");
+    my $ins = $dbh->prepare_cached("INSERT INTO entries (" . join(", ", @cols) . ") VALUES (" . join(", ", ('?') x @cols) . ")");
 
     return ( $ins, @cols );
 }
@@ -479,7 +515,7 @@ sub prepare_update {
 
     my ( $id, @cols ) = @{ $self->_ordered_columns };
 
-    my $upd = $dbh->prepare("UPDATE entries SET " . join(", ", map { "$_ = ?" } @cols) . " WHERE $id = ?");
+    my $upd = $dbh->prepare_cached("UPDATE entries SET " . join(", ", map { "$_ = ?" } @cols) . " WHERE $id = ?");
 
     return ( $upd, @cols, $id );
 }
@@ -566,7 +602,7 @@ sub get {
             my $batch_size = $self->batch_size || scalar(@$ids);
 
             while ( my @batch_ids = splice(@ids_copy, 0, $batch_size) ) {
-                my $sth = $dbh->prepare_cached("SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+                my $sth = $self->prepare_select($dbh, "SELECT id, data FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
                 $sth->execute(@batch_ids);
 
                 $sth->bind_columns( \my ( $id, $data ) );
@@ -673,6 +709,8 @@ sub delete {
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
 
+        my $g = $self->schema->txn_scope_guard;
+
         my $batch_size = $self->batch_size || scalar(@ids);
 
         my @ids_copy = @ids;
@@ -688,6 +726,8 @@ sub delete {
             $sth->execute(@batch_ids);
             $sth->finish;
         }
+
+        $g->commit;
     });
 
     return;
@@ -712,7 +752,7 @@ sub exists {
 
             my @ids_copy = @$ids;
             while ( my @batch_ids = splice @ids_copy, 0, $batch_size ) {
-                my $sth = $dbh->prepare_cached("SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
+                my $sth = $self-> prepare_select ( $dbh, "SELECT id FROM entries WHERE id IN (" . join(", ", ('?') x @batch_ids) . ")");
                 $sth->execute(@batch_ids);
 
                 $sth->bind_columns( \( my $id ) );
@@ -776,7 +816,7 @@ sub _sth_stream {
 
     $self->dbh_do(sub {
         my ( $storage, $dbh ) = @_;
-        my $sth = $dbh->prepare($sql); # can't prepare cached, we don't know when it will be done
+        my $sth = $self->prepare_select($dbh, $sql);
 
         $sth->execute(@bind);
 
@@ -1146,6 +1186,16 @@ argument just before connecting.
 If you need to modify the schema in some way (adding indexes or constraints)
 this is where it should be done.
 
+=item for_update
+
+If true (the defaults), will cause all select statement to be issued with a
+C<FOR UPDATE> modifier on MySQL, Postgres and Oracle.
+
+This is highly reccomended because these database provide low isolation
+guarantees as configured out the box, and highly interlinked graph databases
+are much more susceptible to corruption because of lack of transcational
+isolation than normalized relational databases.
+
 =item sqlite_sync_mode
 
 If this attribute is set and the underlying database is SQLite, then
